"""This file contains groupings and mappings of codes used by stats.
Maybe these should be in the database?
"""


# 5 ethnicity group mapping
ETHNIC_GROUP_MAP = {
    "A": "White",
    "B": "White",
    "C": "White",
    "D": "Mixed",
    "E": "Mixed",
    "F": "Mixed",
    "G": "Mixed",
    "H": "Asian",
    "J": "Asian",
    "K": "Asian",
    "L": "Asian",
    "M": "Black",
    "N": "Black",
    "P": "Black",
    "R": "Asian",
    "S": "Other",
    "Z": "Not Stated",
}

# NHS digital gender map
GENDER_GROUP_MAP = {"1": "Male", "2": "Female", "9": "Indeterminate", "X": "Unknown"}
